(*----------------------------------------------------------------------
Compiler Generator Coco/R,
Copyright (c) 1990, 2004 Hanspeter Moessenboeck, University of Linz
extended by M. Loeberbauer & A. Woess, Univ. of Linz
with improvements by Pat Terry, Rhodes University

This program is free software; you can redistribute it and/or modify it 
under the terms of the GNU General Public License as published by the 
Free Software Foundation; either version 2, or (at your option) any 
later version.

This program is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details.

You should have received a copy of the GNU General Public License along 
with this program; if not, write to the Free Software Foundation, Inc., 
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

As an exception, it is allowed to write an extension of Coco/R that is
used as a plugin in non-free software.

If not otherwise stated, any source code generated by Coco/R (other than 
Coco/R itself) does not fall under the GNU General Public License.
----------------------------------------------------------------------*)


open Tc;


open System;
open System.IO;
open System.Collections.Generic;
open Scanner;


type Errors =
class
val mutable count:int
val mutable errorStream:System.IO.TextWriter
val mutable errMsgFormat:string


	new() = {
		count=0;
		errorStream = Console.Out;
		errMsgFormat = "-- line {0} col {1}: {2}"; //0->line, 1->column, 2->text
	}
	
	member x.SynErr ((line:int), (col:int), (n:int)) =
		let s = 
		match n with
			|0 ->  "EOF expected"
			|1 ->  "IDENT expected"
			|2 ->  "In expected"
			|3 ->  "Fl expected"
			|4 ->  "SET expected"
			|5 ->  "LET expected"
			|6 ->  "IN expected"
			|7 ->  "PRINT expected"
			|8 ->  "DOT expected"
			|9 ->  "PLUS expected"
			|10 ->  "MINUS expected"
			|11 ->  "TIMES expected"
			|12 ->  "DIV expected"
			|13 ->  "EQ expected"
			|14 ->  "LP expected"
			|15 ->  "RP expected"
			|16 ->  "??? expected"
			|17 ->  "invalid Instr"
			|18 ->  "invalid Expr"
			|19 ->  "invalid MulOp"
			|20 ->  "invalid AddOp"
			|21 ->  "invalid Term"
			|22 ->  "invalid Term"
			|23 ->  "invalid Var"

			| _ -> "error " ^ (string_of_int n)
		in
			x.errorStream.WriteLine(x.errMsgFormat,line,col,s);
			x.count<- x.count+1;
	

	member x.SemErr ((line:int), (col:int), (s:string)) =
		x.errorStream.WriteLine(x.errMsgFormat,line,col,s);
		x.count<- x.count+1;
	

	member x.SemErr (s:string) =
		x.errorStream.WriteLine(s);
		x.count<- x.count+1;
	
	member x.Warning((line:int), (col:int), (s:string)) =
		x.errorStream.WriteLine(x.errMsgFormat,line,col,s);
		
	
	member x.Warning(s:string) =
		x.errorStream.WriteLine(s);
	
end // Errors


type parser =
class
	val _EOF : int;
	val _IDENT : int;
	val _In : int;
	val _Fl : int;
	val _SET : int;
	val _LET : int;
	val _IN : int;
	val _PRINT : int;
	val _DOT : int;
	val _PLUS : int;
	val _MINUS : int;
	val _TIMES : int;
	val _DIV : int;
	val _EQ : int;
	val _LP : int;
	val _RP : int;
	val maxT :int;

	val T:bool;
	val x:bool;
	val minErrDist:int;
	val mutable scanner:scanner;
	val mutable errors:Errors;
	val mutable t:Token; //last recognized token
	val mutable la:Token; //lookahead token
	val mutable errDist:int;


	val plus :int ;
	val minus :int ;
	val times :int ;
	val div :int ;
	val equ :int ;


	// types
	val undef :int ;
	val integer :int;
	val real :int ;

	// object kinds
	val print :int ;

	// opcodes
	val ADD :int ;
	val SUB :int ;
	val MUL :int ;
	val DIV :int ;
	val FMUL :int;
	val FADD :int;
	val FSUB :int;
	val FDIV :int;
	val WRITEI :int;
	val WRITEF :int;
	val POP :int;
	val START :int;
	val ERR:int;
	val STOP:int;
	val STOREG:int;
	val STOREL:int;
	val DUP:int;
	val PUSHI:int;
	val PUSHF:int;
	val PUSHG:int;
	val PUSHL:int;
	val PUSHN:int;
   
	val tc : TC;





	new(scan:scanner)= {


		tc= new TC();

		plus = 0;
		minus = 1;
		times = 2;
		div = 3;
		equ = 4;


		// types
		undef = 0;
		integer = 1;
		real = 2;

		// object kinds
		print = 1;

		// opcodes
		ADD = 0;
		SUB = 1;
		MUL = 2;
		DIV = 3;
		FMUL = 4;
		FADD = 5;
		FSUB = 6;
		FDIV = 7;
		WRITEI = 8;
		WRITEF = 9;
		POP =10;
		START = 11;
		ERR=12;
		STOP=13;
		STOREG=14;
		STOREL=15;
		DUP=16;
		PUSHI=17;
		PUSHF=18;
		PUSHG=19;
		PUSHL=20;
		PUSHN=21;
  
  
		_EOF = 0;
		_IDENT = 1;
		_In = 2;
		_Fl = 3;
		_SET = 4;
		_LET = 5;
		_IN = 6;
		_PRINT = 7;
		_DOT = 8;
		_PLUS = 9;
		_MINUS = 10;
		_TIMES = 11;
		_DIV = 12;
		_EQ = 13;
		_LP = 14;
		_RP = 15;
		maxT = 16;
 //initialization of the constants
		scanner =  scan;
		T=true;
		x=false;
		minErrDist=2;
		errors= new Errors();
		t= new Token (0,0,0,0,"dummy",None);
		la= new Token (0,0,0,0,"dummy",None);
		errDist=2;
	}


	member x.SynErr(n:int) = 
		if (x.errDist >= x.minErrDist) then x.errors.SynErr(x.la.line, x.la.col, n);
		x.errDist <- 0;
	

	member x.SemErr(msg:string) =
		if (x.errDist >= x.minErrDist) then x.errors.SemErr(x.t.line, x.t.col, msg);
		x.errDist <- 0;
	
	
	member x.Get() =
		let bboll = ref true in
		while !bboll do
			x.t <- x.la;
			x.la <-x.scanner.Scan();
			if(x.la.kind <= x.maxT)
			then(
				x.errDist <- x.errDist +1;
				bboll:=false;
			);
			
			if(!bboll) then x.la <- x.t;
		done;
	
	
	
	member x.Expect (n:int) =
		if (x.la.kind = n) then x.Get() else  x.SynErr(n); 
	
	
	member x.StartOf (s:int) = x.set(s, (x.la.kind))
	
	
		member x.ExpectWeak ((n:int), (follow:int) ) =
		if (x.la.kind = n) then x.Get()
		else (
			x.SynErr(n);
			while (not(x.StartOf(follow)))do x.Get(); done;
		)
	
	
	member x.WeakSeparator((n:int), (syFol:int), (repFol:int)) =
		let kind = ref x.la.kind in
		if (!kind = n) then ( x.Get(); true )
		else 
			if (x.StartOf(repFol)) then false
			else(
				x.SynErr(n);
				while(not(x.set(syFol, !kind) || x.set(repFol, !kind) || x.set(0, !kind))) do 
					x.Get(); 
					kind:= x.la.kind;
				done;
				x.StartOf(syFol);
			)
	
	
	
	member x.Arith () =
		x.Instr();
		while (x.la.kind = 4 || x.la.kind = 7) do 
			x.Instr();
		 done;
		x.Expect(0);
		x.tc.setImprimir(x.PUSHN, Convert.ToString(x.tc.globais), 0);
		x.tc.setImprimir(x.START, "", 1);
		x.tc.setImprimir(x.PUSHN, Convert.ToString(x.tc.cont), 2);
		x.tc.setImprimir(x.STOP, Convert.ToString(x.tc.locais), x.tc.contador-3);
		x.tc.contador <- x.tc.contador -4;
		        

	member x.Instr () =
		let name = ref " " and tipe = ref 0 and b = ref (new Objectos()) in
		
			if x.tc.cont = 0 then x.tc.cont <- x.tc.locais;
			x.tc.aObjectosLocais <- Array.zero_create 1;
			x.tc.locais<-0;
		          
		if (x.la.kind = 4) then (
			x.Get();
			x.Ident(name);
			b.contents.setObjectosNome(!name); 
			//TC.setGlobais(); 
			                   
			x.Expect(13);
			x.Expr(tipe);
			b.contents.setObjectosTipo(!tipe); 
			
						x.tc.setObjectosGlobais(!b);
					
			x.Expect(8);
			let lugar = x.tc.getObjIndex(x.tc.aObjectosGlobais, name.contents) in
			x.tc.setImprimir(x.STOREG, lugar.ToString() , x.tc.contador);   
			      
		) else if (x.la.kind = 7) then ( 
			x.Get();
			x.Expr(tipe);
			if(!tipe = x.integer) then  x.tc.setImprimir(8, "", (x.tc.contador))
			else if (!tipe = x.	real) then x.tc.setImprimir(9, "", x.tc.contador);
			                   
			x.Expect(8);
		) else x.SynErr(17);

	member x.Ident (name : string byref) =
		x.Expect(1);
		name <- x.t.value;  

	member x.Expr (tipe:int byref) =
		let name = ref "" and tp= ref x.undef and op = 0 and b = new Objectos() in
		if (x.StartOf(1)) then (
			x.Calculo(tp);
			tipe <- !tp;
		) else if (x.la.kind = 5) then ( 
			x.Get();
			x.Ident(name);
			x.Expect(13);
			x.Expr(tp);
			tipe <- !tp;
			let lugar = ref 0 in
			b.setObjectosNome(!name);
			b.setObjectosTipo(tipe); 
			x.tc.setObjectosLocais(b);
			 //x.tc.aObjectosLocais.(x.tc.aObjectosLocais.Length)<-b;
			
			if(x.tc.getObjIndex(x.tc.aObjectosLocais, name.contents) <> -1)
			then(
			     lugar := x.tc.getObjIndex(x.tc.aObjectosLocais, name.contents);
			     x.tc.setImprimir(x.STOREL, lugar.contents.ToString() , x.tc.contador);
			 );
			                                    
			x.Expect(6);
			x.Expr(tp);
			tipe <- !tp;
			//if(x.tc.locais < x.tc.aObjectosLocais.Length) then ignore(x.tc.setLocais());
			
		) else x.SynErr(18);

	member x.Calculo (tipe: int byref) =
		let tp = ref x.undef and tipe1 = ref 0 and tipe2 = ref 0 and operador = ref x.undef and op = ref 0 in
		x.Term(tp);
		tipe <- !tp;
		while (x.la.kind = 9 || x.la.kind = 10) do 
			x.AddOp(op);
			x.Term(tipe1);
			if (tipe <> !tipe1 ) then x.errors.SemErr(x.la.line, x.la.col, "tipos diferentes"); //  Err("tipos diferentes");
			
			if(tipe = x.integer)then(
			                      if(!op=x.minus) then operador:=x.SUB ;
			                   if(!op=x.plus) then operador:=x.ADD;
			                  );
			               
			               if(tipe=x.real)then (
			                if(!op=x.minus) then operador:=x.FSUB ;
			                   if(!op=x.plus) then operador:=x.FADD;
			                   );
			               x.tc.setImprimir(!operador, "" , x.tc.contador); 
			                          
		 done;

	member x.MulOp (op: int byref) =
		op <- -1;
		if (x.la.kind = 11) then (
			x.Get();
			op<-x.times;
		) else if (x.la.kind = 12) then ( 
			x.Get();
			op<-x.div;
		) else x.SynErr(19);

	member x.AddOp (op: int byref ) =
		op <- -1;
		if (x.la.kind = 9) then (
			x.Get();
			op <- x.plus;
		) else if (x.la.kind = 10) then ( 
			x.Get();
			op <- x.minus;
		) else x.SynErr(20);

	member x.Term (tipe : int byref) =
		let tp = ref x.undef and tipe1 = ref 0 and tipe2=ref 0 and operator = ref x.undef and op=ref 0 in
		tipe<-x.undef; 
		if (x.la.kind = 1 || x.la.kind = 2 || x.la.kind = 3) then (
			x.Var(tp);
			tipe<- !tp;
		) else if (x.la.kind = 10) then ( 
			x.Negacao(tp);
			tipe<- !tp;
		) else if (x.la.kind = 14) then ( 
			x.Parentesis(tp);
			tipe<- !tp;
		) else x.SynErr(21);
		while (x.la.kind = 11 || x.la.kind = 12) do 
			x.MulOp(op);
			if (x.la.kind = 1 || x.la.kind = 2 || x.la.kind = 3) then (
				x.Var(tipe1);
				if (tipe <> !tipe1 ) then  x.errors.SemErr(x.la.line, x.la.col, "tipos diferentes"); // Err("tipos diferentes");
				
			) else if (x.la.kind = 14) then ( 
				x.Parentesis(tipe2);
				if(tipe <> !tipe2) then	x.errors.SemErr(x.la.line, x.la.col, "tipos diferentes"); //Err("tipos diferentes");
				
			) else x.SynErr(22);
			if(tipe = x.integer)then(
			if(!op=x.times) then operator :=x.MUL;
			if(!op=x.div) then operator :=x.DIV;
			);
			if(tipe=x.real)then(
			   if(!op=x.times)then operator:=x.FMUL;
			    if(!op=x.div) then operator:=x.FDIV;
			);
			x.tc.setImprimir(!operator, "" , x.tc.contador);    
		 done;

	member x.Var (tipe:int byref) =
		let tp = ref x.undef and name = ref "" in
		tipe <- x.undef; 
		if (x.la.kind = 3) then (
			x.Get();
			tipe <- x.real;
			x.tc.setImprimir(x.PUSHF, x.t.value, x.tc.contador );
			     
		) else if (x.la.kind = 2) then ( 
			x.Get();
			tipe <- x.integer; 
			//printf "entro nest push i? %s local %d\n" x.t.value x.tc.contador;
			x.tc.setImprimir(x.PUSHI, x.t.value, x.tc.contador);
			        
		) else if (x.la.kind = 1) then ( 
			x.Ident(name);
			let lugar = ref 0 in
			if((x.tc.aObjectosLocais <> null) &&  (x.tc.getObjIndex(x.tc.aObjectosLocais, name.contents) <> -1))
			then(
			    tipe <- x.tc.getObjTipo(x.tc.aObjectosLocais, name.contents);
			    lugar := x.tc.getObjIndex(x.tc.aObjectosLocais, name.contents);
			    x.tc.setImprimir(x.PUSHL, lugar.contents.ToString() , x.tc.contador);
			 )
			else 
			    if((x.tc.aObjectosGlobais <> null) &&  (x.tc.getObjIndex(x.tc.aObjectosGlobais, name.contents) <> -1))
			    then (
			        tipe<- x.tc.getObjTipo(x.tc.aObjectosGlobais, name.contents);
			        lugar := x.tc.getObjIndex(x.tc.aObjectosGlobais, name.contents);
			        x.tc.setImprimir(x.PUSHG, lugar.contents.ToString() , x.tc.contador); 
			     )
			                      
		) else x.SynErr(23);

	member x.Negacao (tipe:int byref) =
		let tp = ref 0 and lugar =ref 0 and name1 = "loc" and b = new Objectos() in
		x.Expect(10);
		x.Parentesis(tp);
		tipe <- !tp;
		
		b.setObjectosNome(name1);
		b.setObjectosTipo(tipe); 
		x.tc.setObjectosLocais(b);
		//x.tc.aObjectosLocais.(x.tc.aObjectosLocais.Length)<- b ;
		x.tc.setImprimir(x.PUSHI, "0" , x.tc.contador);                 
		if(x.tc.getObjIndex(x.tc.aObjectosLocais, name1) <> -1)
		then(
			  lugar := x.tc.getObjIndex(x.tc.aObjectosLocais, name1);
			  x.tc.setImprimir(x.STOREL , lugar.contents.ToString() , x.tc.contador);
			
			if(tipe= x.integer)
			then x.tc.setImprimir(x.SUB, "" , x.tc.contador);
			
			if(tipe=x.real)then
			x.tc.setImprimir(x.FSUB, "" , x.tc.contador);
		)
		

	member x.Parentesis (tipe: int byref) =
		let op = 0  and tp = ref x.undef in 
		x.Expect(14);
		x.Expr(tp);
		tipe<- !tp;
		x.Expect(15);



	member x.Parse() =
		x.la <- new Token (0,0,0,0,"dummy",None);
		x.Get();
		x.Arith();

		x.Expect(0);
	
	
	member x.Find((n:int), (bf:bool array) ) =
		try
			let t =	(bf.GetValue(Int64.of_int n):?>bool) in
			t
		with _ -> false
	
	
	member x.set((a:int),(b:int))= 
		match a with
		|0 -> let t = [|x.T;x.x;x.x;x.x; x.x;x.x;x.x;x.x; x.x;x.x;x.x;x.x; x.x;x.x;x.x;x.x; x.x;x.x|] in x.Find(b,t)
		|1 -> let t = [|x.x;x.T;x.T;x.T; x.x;x.x;x.x;x.x; x.x;x.x;x.T;x.x; x.x;x.x;x.T;x.x; x.x;x.x|] in x.Find(b,t)

		|_ -> false

end

